"""
Super Collections

Their purpose is to turn complex input such as json or YAML files into
Python objects accessible with attributes, and self documented.

The general idea is that those structured files are combinations
of lists and dictionaries.

(C) Laurent Franceschetti 2024
"""
import datetime
import json
import inspect
from typing import Any


import hjson

# -------------------------------------
# Low-level fixtures
# -------------------------------------
from collections import UserDict

DICT_TYPES = dict, UserDict

class CustomEncoder(json.JSONEncoder):
    """
    Custom encoder for JSON serialization.
    Used for debugging purposes.
    """
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, UserDict):
            # for objects used by some packages
            return dict(obj)

        elif inspect.isfunction(obj):
            return f"Function: %s %s" % (inspect.signature(obj),
                                        obj.__doc__)
        try:
            return super().default(obj)
        except TypeError:
            print(f"CANNOT INTERPRET {obj.__class__}")
            return str(obj)


# -------------------------------------
# Collections
# -------------------------------------
class SuperDict(dict):
    """
    A dictionary with key accessibles as properties
    (with the dot notation)

    a['foo'] <=> a.foo

    As a rule, the Superdict will expose as properties
    all keys that:

    1. Are valid identifiers
    2. Are not a standard property or method of the dict, 
        class notably:
        attributes, clear, copy, fromkeys, get, items,
        keys, pop, popitem, setdefault, update, values

    Lists in a SuperDict are converted into SuperLists, whose elements
    are in turn converted, etc...
    """

    def __init__(self, *args, **kwargs):
        # Call the superclass's __init__ method
        super().__init__(*args, **kwargs)
        self.__post_init__()

    def __post_init__(self):
        "Recursively transform sub-dictionary"
        for key, value in self.items():
            if isinstance(value, SUPER_TYPES):
                pass
            elif isinstance(value, DICT_TYPES):
                self[key] = SuperDict(value)
            elif isinstance(value, list):
                self[key] = SuperList(value)

    def __getattr__(self, name:str):
        "Allow dot notation on reading"
        ERR_MSG = "Cannot find attribute '%s'" % name
        # if name.startswith('_'):
        #     raise AttributeError(ERR_MSG)
        try:
            return self[name]
        except KeyError:
            raise AttributeError(ERR_MSG)



    def properties(self):
        """
        Generate the valid properties
        (the dictionary keys that qualify as Python identifiers
        and are not callables)
        """
        return (item for item in self.keys() 
                if isinstance(item, str)
                    and item.isidentifier()
                    and not callable(getattr(self, item)))
    

    
    def __dir__(self):
        "List all attributes (for autocompletion, etc.)"
        return super().__dir__() + list(self.properties())
    
    # -------------------------------------
    # Output
    # -------------------------------------


    def __setattr__(self, name, value):
        "Allow dot notation on writing"
        # ERR_MSG = "Cannot assign an attribute starting with _ ('%s')" % name
        # if name.startswith('_'):
        #     raise AttributeError(ERR_MSG)     
        self[name] = value


    def update(self, other:dict):
        """
        Update the SuperDict with another.

        If necessary the other dictionary is converted into a SuperDict
        """
        if not isinstance(other, SuperDict):
            other = SuperDict(other)
        return super().update(other)
    
    
    # -------------------------------------
    # Output
    # -------------------------------------
    
    def to_json(self):
        """
        Convert to json.

        It does not have any claim of fitness for any
        particular purpose, except showing what's in structure,
        for string output.
        """
        return json.dumps(self, cls=CustomEncoder)
    
    def to_hjson(self):
        """
        Convert to hjson.

        It does not have any claim of fitness for any
        particular purpose, except showing what's in structure,
        for string output.
        """
        python_dict = json.loads(self.to_json())
        return hjson.dumps(python_dict)


    def __str__(self):
        "Print a superdict"
        return self.to_hjson()
    
    def __rich__(self):
        "Print a superdict (for rich)"
        r = [f"[bold red]{self.__class__.__name__}:[/]"]
        r.append(self.to_hjson())
        return("\n".join(r))       




class SuperList(list):
    """
    A list that supports the SuperDict,
    to allow recursion within complex structures
    """

    def __init__(self, *args, **kwargs):
        # Call the superclass's __init__ method
        super().__init__(*args, **kwargs)
        self.__post_init__()

    def __post_init__(self):
        "Recursively transform sub-list"
        for index, value in enumerate(self):
            if isinstance(value, SUPER_TYPES):
                pass
            elif isinstance(value, DICT_TYPES):
                self[index] = SuperDict(value)
            elif isinstance(value, list):
                self[index] = SuperList(value)



    # -------------------------------------
    # Modify
    # -------------------------------------

    def extend(self, l):
        "Extend the list with another one (transforms it first in SuperList)"
        l = SuperList(l)
        super().extend(l)

    def __add__(self, l):
        "Addition with another list"
        l = SuperList(l)
        return SuperList(super().__add__(l))
    # -------------------------------------
    # Output
    # -------------------------------------
    
    def to_json(self):
        """
        Convert to json.

        It does not have any claim of fitness for any
        particular purpose, except showing what's in structure,
        for string output.
        """
        return json.dumps(self, cls=CustomEncoder)
    
    def to_hjson(self):
        """
        Convert to hjson.

        It does not have any claim of fitness for any
        particular purpose, except showing what's in structure,
        for string output.
        """
        python_dict = json.loads(self.to_json())
        return hjson.dumps(python_dict)


    def __str__(self):
        "Print a superdict"
        return self.to_hjson()
    
    def __rich__(self):
        "Print a superdict (for rich)"
        r = [f"[bold red]{self.__class__.__name__}:[/]"]
        r.append(self.to_hjson())
        return("\n".join(r))     

SUPER_TYPES = SuperDict, SuperList


"""
Testing Super Collections with a pandoc-generated file
(https://pandoc.org)

The purpose is to show a practical application.

It could be done directly on lists and dictionary, but it would
be awkward. The code is much more legible and maintainable.


(C) Laurent Franceschetti 2024
"""

import os
import json

import pytest


from super_collections import SuperDict, SuperList

# -------------------------------------
# Constants
# -------------------------------------

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

PANDOC_FILE = 'test_pandoc.json'


# -------------------------------------
# Code
# -------------------------------------

def strip_format(sl: SuperList) -> str:
    "Print pandoc text from a super list"
    if not isinstance(sl, SuperList):
        raise ValueError(f"{sl} not a list")
    r = []
    for element in sl:
        if isinstance(element, (str, int)):
            pass
        elif isinstance(element, list):
            r.append(strip_format(element))
        elif element.t == 'Str':
            r.append(element.c)
        elif element.t == 'Space':
            r.append(' ')
        elif isinstance(element.c, list):
            r.append(strip_format(element.c))
    return ''.join(r)

def test_read_json_file():
    """
    Read the JSON file using the function and convert it into a SuperDict
    """
    filename = os.path.join(CURRENT_DIR, PANDOC_FILE)
    with open(filename, 'r') as file:
        data = json.load(file)
    document = SuperDict(data)

    print('\n----')
    lines = []
    for block in document.blocks:
        # change the header level
        if block.t == 'Header':
            block.c[0] += 1
        lines.append(strip_format(block.c))
    result = '\n'.join(lines)
    print(result)
    assert "fairly heavy" in result
    assert "Title level 2" in result






import os
import json


import pytest


from super_collections import SuperDict, SuperList

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

SOLAR_SYSTEM = 'solar_system.json'


def test_dict():
    """
    Test Superdict
    """
    DATA = """{
        "name": "Mercury",
        "size": 4880,
        "mass": 3.30e23,
        "orbit": 57900000,
        "orbit_from": "Sun",
        "length_of_year": 88,
        "moons": [],
        "foo bar": 50,
        "items": "foobar"
      }"""
    
    mercury = json.loads(DATA)
    mercury = SuperDict(mercury)

    # normal diction
    assert 'name' in mercury
    assert 'size' in mercury

    # attributes
    assert mercury.name == 'Mercury'
    assert mercury.size == 4880
    assert mercury.orbit_from == "Sun"

    # discrovery
    assert 'name' in mercury.properties()
    assert 'name' in dir(mercury)
    assert 'orbit_from' in dir(mercury)

    # invalid identifier
    invalid_id = 'foo bar'
    assert invalid_id in mercury
    assert invalid_id not in dir(mercury)
    assert mercury[invalid_id] == 50 # but reachable as a dict key

    # not a valid property (is masked by the standard method `items()`)
    assert mercury['items'] == 'foobar'
    assert 'items' not in mercury.properties()
    assert mercury.items != 'foobar'
    assert callable(mercury.items)
    assert ('size', 4880) in mercury.items()
 



def test_update_list():
    "Update super lists"

    l = SuperList(({'foo': 5, 'bar': 5}, 'baz'))
    assert l[0].foo == 5

    l2 = [6, {'barbaz': 45}]
    l.extend(l2)

    last = l[-1]
    assert last.barbaz == 45


    l3 = ["hello", {'foobar': 30}]

    # add a list to a SuperList -> Superlist
    r1 = l + l3
    assert isinstance(r1, SuperList)
    last = r1[-1]
    assert last.foobar == 30 # this is a SuperDict

    # CAUTION: add a SuperList to a List -> list
    r2 = l3 + l
    assert not isinstance(r2, SuperList)
    with pytest.raises(AttributeError):
        second = r2[1]
        second.foobar # this is NOT a SuperDict


def test_read_json_file():
    """
    Read the JSON file using the function and convert it into a SuperDict
    """
    filename = os.path.join(CURRENT_DIR, SOLAR_SYSTEM)
    with open(filename, 'r') as file:
        data = json.load(file)
    solar_system = SuperDict(data)

    # attributes are recursively available, including with lists:
    units = solar_system.units
    assert units.size == 'km'
    for planet in solar_system.planets:
        assert isinstance(planet, SuperDict)
        print (planet.name, planet.size)
        for moon in planet.moons:
            print(f"- {moon.name}", moon.size)

    ADDITIONAL = [{'name': 'Foo', 'size': 45895},
                  {'name': 'Bar', 'size': 59383}]
    new_planet_list = solar_system.planets + ADDITIONAL
    assert isinstance(new_planet_list, SuperList)
    last_planet = new_planet_list[-1]
    assert last_planet.name == 'Bar'
    assert last_planet.size == 59383
    with pytest.raises(AttributeError):
        last_planet.orbit
    last_planet.orbit = 5930590
    assert last_planet.orbit == 5930590 

if __name__ == "__main__":
    pytest.main()
